/* gscomp : compresses files
 * Copyright (C) 2004  Alexandre Boeglin <alex@boeglin.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <netinet/in.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>

typedef unsigned char u8;
typedef unsigned int u32;
typedef unsigned long int u64;

int main(int argc, char **argv){

  char ident_string[9] = "ngc-comp";
  int fd_in, fd_out, i, error_test;
  struct stat stat_in;

  u32 block_size, block_number, header_size, compressed_size, actual_read_size, write_offset;
  u64 actual_compressed_size;

  u8 *buffer_in, *buffer_out;
  u32 *header;

  // check args
  if(argc !=4)
    exit(1);

  // get block size
  block_size = atoi(argv[1]);
  if(block_size == 0)
    exit(2);

  // open files
  fd_in = open(argv[2], O_RDONLY);
  if(fd_in == -1)
    exit(3);

  fd_out = open(argv[3], O_WRONLY | O_CREAT | O_TRUNC, S_IRUSR | S_IWUSR);
  if(fd_out == -1)
    exit(4);

  // get file size and calculate block number
  error_test = fstat(fd_in, &stat_in);
  if(error_test == -1)
    exit(5);

  block_number = stat_in.st_size / block_size;
  if(stat_in.st_size % block_size)
    block_number++;

  // header will only be copied to file in the end, when values are filled
  header_size = sizeof(u32) * (2 * block_number + 4);
  header = malloc(header_size);
  if(header == NULL)
    exit(6);

  memcpy(header, ident_string, 8);
  header[2] = htonl(block_size);
  header[3] = htonl(block_number);

  // prepare to write
  error_test = lseek(fd_out, header_size, SEEK_SET);
  if(error_test == -1)
    exit(7);

  write_offset = header_size;
  compressed_size = compressBound(block_size);

  buffer_in = malloc(block_size);
  if(buffer_in == NULL)
    exit(8);
  buffer_out = malloc(compressed_size);
  if(buffer_out == NULL)
    exit(9);

  // compression loop
  for(i = 0; i < block_number; i++){
    // read data
    actual_read_size = read(fd_in, buffer_in, block_size);
    if(actual_read_size == -1)
      exit(10);

    // compress data
    actual_compressed_size = compressed_size;
    error_test = compress2(buffer_out, &actual_compressed_size, buffer_in, actual_read_size, Z_DEFAULT_COMPRESSION);
    if(error_test != Z_OK)
      exit(11);

    // update header
    header[2 * i + 4] = htonl(write_offset);
    header[2 * i + 5] = htonl(actual_compressed_size);

    // write data
    error_test = write(fd_out, buffer_out, actual_compressed_size);
    if(error_test == -1)
      exit(12);

    write_offset += actual_compressed_size;
  }
  // write header
  error_test = lseek(fd_out, 0, SEEK_SET);
  if(error_test == -1)
    exit(13);

  error_test = write(fd_out, header, header_size);
  if(error_test == -1)
    exit(14);

  fprintf(stderr, "File compressed without errors!\n");
  return 0;
}
