/* gsuncomp : uncompresses files
 * Copyright (C) 2004  Alexandre Boeglin <alex@boeglin.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <netinet/in.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <zlib.h>

typedef unsigned char u8;
typedef unsigned int u32;
typedef unsigned long int u64;

int main(int argc, char **argv){

  char ident_string[9] = "ngc-comp", comp_string[9] = "12345678";
  int fd_in, fd_out, i, error_test;

  u32 block_size, block_number, header_size, actual_read_size;
  u64 actual_uncompressed_size;
  u8 *buffer_in, *buffer_out;
  u32 *header;

  // check args
  if(argc != 3)
    exit(1);

  // open files
  fd_in = open(argv[1], O_RDONLY);
  if(fd_in == -1)
    exit(2);

  fd_out = open(argv[2], O_WRONLY | O_CREAT, S_IRUSR | S_IWUSR);
  if(fd_out == -1)
    exit(3);

  // check ident string
  error_test = read(fd_in, comp_string, 8);
  if(error_test == -1)
    exit(4);
  if(memcmp(ident_string, comp_string, 8) != 0)
    exit(5);

  // get block size and block number
  error_test = read(fd_in, &block_size, sizeof(block_size));
  if(error_test == -1)
    exit(6);
  block_size = ntohl(block_size);

  error_test = read(fd_in, &block_number, sizeof(block_number));
  if(error_test == -1)
    exit(7);
  block_number = ntohl(block_number);

  // get header
  header_size = sizeof(u32) * (2 * block_number);
  header = malloc(header_size);
  if(header == NULL)
    exit(8);

  error_test = read(fd_in, header, header_size);
  if(error_test == -1)
    exit(9);

  for(i = 0; i < (2 * block_number); i++)
    header[i] = ntohl(header[i]);

  // prepare to write
  buffer_in = malloc(block_size * 2);
  if(buffer_in == NULL)
    exit(10);

  buffer_out = malloc(block_size);
  if(buffer_out == NULL)
    exit(11);

  // uncompression loop
  for(i = 0; i < block_number; i++){
    // read data
    actual_read_size = read(fd_in, buffer_in, header [2 * i + 1]);
    if(actual_read_size == -1)
      exit(12);

    // uncompress data
    actual_uncompressed_size = block_size;
    error_test = uncompress(buffer_out, &actual_uncompressed_size, buffer_in, actual_read_size);
    if(error_test != Z_OK)
      exit(13);

    // write data
    error_test = write(fd_out, buffer_out, actual_uncompressed_size);
    if(error_test == -1)
      exit(14);
  }
  fprintf(stderr, "File uncompressed without errors!\n");
  return 0;
}
